(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Conditional_Rewriting_Impl
imports
  Conditional_Rewriting
  QDP_Framework_Impl
begin

type_synonym ('f, 'v) crules = "('f, 'v) crule list"

fun shows_conditions :: "('f :: show, 'v :: show) rule list \<Rightarrow> bool \<Rightarrow> shows"
where
  "shows_conditions [] _ = id" |
  "shows_conditions (x # xs) b = (shows (if b then '' | '' else '', '') +@+ 
    shows_rule' shows shows '' ->* '' x +@+ shows_conditions xs False)"

definition shows_crule :: "('f :: show, 'v :: show) crule \<Rightarrow> shows"
where
  "shows_crule cr = (shows_rule (fst cr) +@+ shows_conditions (snd cr) True)"

definition check_type3 :: "('f\<Colon>show, 'v\<Colon>show) crules \<Rightarrow> shows check"
where
  "check_type3 R = do {
    check_allm (\<lambda>cr. check_subseteq (vars_term_list (snd (fst cr))) (vars_term_list (fst (fst cr)) @ vars_trs_list (snd cr))
     <+? (\<lambda>x. ''variable '' +#+ shows x
        +@+ '' occurs only in right-hand side of rule '' +#+ shows_crule cr +@+ shows_nl)) R
  } <+? (\<lambda>e. ''the CTRS is not of type 3'' +#+ shows_nl +@+ e)"

lemma check_type3 [simp]:
  "isOK (check_type3 R) = type3 (set R)"
  unfolding check_type3_def type3_def by force

definition X_impl :: "('f, 'v) crule \<Rightarrow> nat \<Rightarrow> 'v list"
 where "X_impl cr i = 
  concat (vars_term_list (fst (fst cr)) # (map (vars_term_list \<circ> snd) (take i (snd cr)) ))"

lemma X_impl: 
 "set (X_impl cr i) = X_vars cr i" 
 unfolding X_impl_def X_vars_def by force

definition Y_impl :: "('f, 'v) crule \<Rightarrow> nat \<Rightarrow> 'v list"
 where "Y_impl cr i = 
  vars_term_list (snd (fst cr)) @ vars_term_list (snd ((snd cr) ! i)) @ (vars_trs_list (drop i (snd cr)))"

lemma Y_impl: 
 "set (Y_impl cr i) = Y_vars cr i" 
 unfolding Y_impl_def Y_vars_def by force


definition check_dctrs :: "('f\<Colon>show, 'v\<Colon>show) crules \<Rightarrow> shows check"
where
  "check_dctrs R = do {
    check_allm 
     (\<lambda>cr.
      check_allm
       (\<lambda>i.
        check_subseteq (vars_term_list (fst (snd cr ! i))) (X_impl cr i)
         <+? (\<lambda>x. ''variable '' +#+ shows x +@+ '' in condition '' +#+ shows_rule (snd cr ! i) +@+ '' of rule '' 
                +#+  shows_crule cr +@+ ''violates DCTRS condition'' +#+ shows_nl)
       ) [0..<length (snd cr)]
     ) R 
  } <+? (\<lambda>e. ''the CTRS is not of type 3'' +#+ shows_nl +@+ e)"

lemma check_dctrs [simp]:
  "isOK (check_dctrs R) = dctrs (set R)"
proof-
 have x:"\<And>k P. (\<forall>x\<in>set [0..<k]. P x) =  (\<forall>x < k. P x) " by auto
 thus ?thesis unfolding check_dctrs_def dctrs_def isOK_update_error isOK_forallM 
  isOK_check_subseteq X_impl x by simp
qed

definition check_wf_ctrs :: "('f\<Colon>show, 'v\<Colon>show) crules \<Rightarrow> shows check"
where
  "check_wf_ctrs R = do {
    check_allm (\<lambda>cr. 
      check (is_Fun (fst (fst cr)))
        (''variable left-hand side in rule '' +#+ shows_crule cr  +@+ shows_nl)
    ) R;
    check_dctrs R;
    check_type3 R
  } <+? (\<lambda>e. ''the CTRS is not well-formed'' +#+ shows_nl +@+ e)"

lemma check_wf_ctrs [simp]:
  "isOK (check_wf_ctrs R) = wf_ctrs (set R)"
  unfolding check_wf_ctrs_def wf_ctrs_def
  by force+

definition funs_crule_list :: "('f, 'v) crule \<Rightarrow> 'f list"
where
  "funs_crule_list r = add_funs_rule (fst r) (funs_trs_list (snd r))"

lemma funs_crule_list_sound [simp]:
  "set (funs_crule_list r) = funs_crule r"
  by (simp add: funs_crule_list_def funs_crule_def)

definition funs_ctrs_list :: "('f, 'v) crule list \<Rightarrow> 'f list"
where
  "funs_ctrs_list trs = concat (map funs_crule_list trs)"

lemma funs_ctrs_list_sound [simp]:
  "set (funs_ctrs_list r) = funs_ctrs (set r)"
  by (simp add: funs_ctrs_list_def funs_ctrs_def)

end

