(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Graph
imports
  Check_Monad
  "$AFP/Show/Show"
begin

type_synonym 'a graph = "'a set \<times> ('a \<times> 'a) set"
type_synonym 'a ipath = "nat \<Rightarrow> 'a"

definition nodes :: "'a graph \<Rightarrow> 'a set"
where "nodes G = fst G"
  

definition
  path_nodes :: "'a ipath \<Rightarrow> 'a set"
where
  "path_nodes p = range p"

definition
  suffix :: "'a ipath \<Rightarrow> 'a ipath \<Rightarrow> bool" ("_ =\<dots> _" 50)
where
  "(p =\<dots> q) = (\<exists> n. \<forall> m. (p (m+n) = q m))"

lemma suffix_trans[trans]: assumes "p =\<dots> q" and "q =\<dots> r" shows "p =\<dots> r"
using assms proof (unfold suffix_def)
  assume "\<exists>n. \<forall>m. p (m + n) = q m" 
  from this obtain n1 where A: "\<forall> m. p (m + n1) = q m" ..
  assume "\<exists>n. \<forall>m. q (m + n) = r m"
  from this obtain n2 where B: "\<forall>m. q (m + n2) = r m" ..
  have " \<forall>m. p (m + (n2 + n1)) = (r m)" 
  proof 
    fix m
    have "m + (n2 + n1) = m + n2 + n1" (is "?x = ?y") by arith
    hence "p (m + (n2 + n1)) = p (m + n2 + n1)" using arg_cong[where x = ?x and y = ?y] by simp
    also from A have "\<dots> = q (m + n2) " ..
    also from B have "\<dots> = r m" ..
    finally show "p (m + (n2 + n1)) = r m" .
 qed
 thus  "\<exists>n. \<forall>m. p (m + n) = r m" ..
qed


(* start of automated check *)

definition
  check_edges :: "('a \<Rightarrow> shows) \<Rightarrow> ('a\<times>'a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> shows check"
where
  "check_edges ss g c d =
     check_allm (\<lambda>n. check_allm (\<lambda>m.
       check (\<not> g (m, n)) (''edge from '' +#+ ss m +@+ '' to '' +#+ ss n)) d) c"

lemma check_edges_sound[simp]:
  "isOK (check_edges ss g c d) = (\<forall>n\<in>set c. (\<forall>m\<in>set d. (\<not> g (m, n))))"
  unfolding check_edges_def
  by auto

definition
  check_graph_decomp
    :: "('a \<Rightarrow> shows) \<Rightarrow> ('a \<times> 'a \<Rightarrow> bool) \<Rightarrow> (bool \<times> 'a list) list \<Rightarrow> shows check"
where
  "check_graph_decomp ss g rcs = do {
     check_pairwise (check_edges ss g) (map snd rcs);
     check_allm (\<lambda>c. check_edges ss g c c) (map snd (filter (\<lambda>rc. \<not> (fst rc)) rcs))
   }"

lemma check_graph_decomp_sound: 
  assumes check: "isOK (check_graph_decomp ss g cs)"
  and path: "\<And> n. g (p n, p (Suc n))"
  and nodes: "path_nodes p \<subseteq> set (concat (map snd cs))" (is "?pn p cs")
  shows "\<exists> q C . p =\<dots> q \<and> C \<in> set (map snd (filter fst cs)) \<and> path_nodes q \<subseteq>  set C" 
proof -
  let ?cs = "map snd cs"
  let ?ncs = "map snd (filter (\<lambda> rc. \<not> (fst rc)) cs)"
  let ?inn = "\<lambda> c. (\<forall> n \<in> set c. \<forall> m \<in> set c. \<not> (g (m,n)))"
  note check = check[unfolded check_graph_decomp_def]
  from check have nr_sccs: "\<forall> c \<in> set ?ncs. ?inn c" (is "?in cs") by auto
  from check have back_e: "\<forall> j < length cs. \<forall> i < j. \<forall> n \<in> set (?cs ! i). \<forall> m \<in> set (?cs ! j). \<not> (g (m,n))" (is "?back cs") by auto
  from path nr_sccs back_e nodes show ?thesis 
  proof (induct cs arbitrary: p)
    case Nil
    thus ?case unfolding path_nodes_def by auto
  next
    case (Cons rc cs)
    let ?cs = "map snd cs"
    let ?ncs = "map snd (filter (Not o fst) cs)"    
    {
      fix q
      assume "p =\<dots> q"
      from this[unfolded suffix_def] obtain n where q: "\<And> m. p (m + n) = q m" by auto        
      hence q: "\<And> m. q m = p (m + n)" by simp
      from Cons(2) have "\<And> n. g (q n, q (Suc n))" unfolding q by simp
    } note path_suffix = this
    show ?case
    proof (cases "\<exists> q. p =\<dots> q \<and> path_nodes q \<subseteq> set (concat ?cs)")
      case True
      from this obtain q where pq: "p =\<dots> q" and qnodes: "?pn q cs" by auto
      from Cons(4) have "?back cs" by force
      have "?in cs" 
      proof
	fix c
	assume "c \<in> set (map snd [r\<leftarrow>cs . \<not> fst r])"
	hence "c \<in> set (map snd [r\<leftarrow>rc # cs . \<not> fst r])" by auto
	with `?in (rc # cs)` show "?inn c" by auto 
      qed
      from path_suffix[OF pq] `?pn q cs` `?in cs` `?back cs` Cons 
      have "\<exists> r C. q =\<dots> r \<and> C \<in> set (map snd (filter fst cs)) \<and> path_nodes r \<subseteq> set C" by auto
      then obtain r C where qr: "q =\<dots> r" and C: "C \<in> set (map snd (filter fst cs))" and r: "path_nodes r \<subseteq> set C" by auto
      from pq qr have p_r: "p =\<dots> r" using suffix_trans by auto
      from C have "C \<in> set (map snd (filter fst (rc # cs)))" by auto
      with p_r r show ?thesis by blast
    next
      case False
      hence ends_not_in_cs: "\<forall> q. p =\<dots> q \<longrightarrow> (\<not> path_nodes q \<subseteq> set (concat ?cs))" by auto
      show ?thesis
      proof (cases rc)
	case (Pair r c)
	show ?thesis 
	proof (cases "\<exists> q. p =\<dots> q \<and> path_nodes q \<subseteq> set c")
	  case True
	  from this obtain q :: "'a ipath" where pq: "p =\<dots> q" and pnodes: "path_nodes q \<subseteq> set c" by auto
	  have r
	  proof (rule ccontr)
	    assume "\<not> r"
	    with Cons Pair have contra: "?inn c" by auto
            from path_suffix[OF pq]
              contra pnodes show False
	      unfolding path_nodes_def by force
	  qed
	  with pq pnodes Pair show ?thesis  by auto
	next
	  case False
	  have "p =\<dots> p" unfolding suffix_def
	  proof
	    show "\<forall> m. p (m + 0) = p m" by auto
	  qed
	  with False have "\<not> path_nodes p \<subseteq> set c" by auto
	  from this obtain i where i: "p i \<notin> set c" unfolding path_nodes_def by auto 
	  obtain q where q: "\<forall> j. p (j + (Suc i)) = q j" by force
	  hence pq: "p =\<dots> q" unfolding suffix_def ..
	  with ends_not_in_cs have "\<not> path_nodes q \<subseteq> set (concat ?cs)" by auto
	  from this obtain j where j: "q j \<notin> set (concat ?cs)" unfolding path_nodes_def by auto
	  let ?k = "Suc (j + i)"
	  from j q have k: "p ?k \<notin> set (concat ?cs)" by auto
	  from i k Pair `?pn p (rc # cs)` have nearly: "p i \<in> set (concat ?cs) \<and> p ?k \<in> set c"
	    unfolding path_nodes_def by (simp, blast)
	  hence "\<exists> k. p k \<in> set (concat ?cs) \<and> p (Suc k) \<in> set c"
	  proof (induct j arbitrary: i)
	    case 0
	    thus ?case by auto
	  next
	    case (Suc j)
	    thus ?case 
	    proof (cases "p (Suc i) \<in> set c")
	      case True
	      with Suc show ?thesis by auto
	    next
	      case False
	      with `?pn p (rc # cs)` Pair have p: "p (Suc i) \<in> set (concat ?cs)" unfolding path_nodes_def by auto
	      have su: "Suc (Suc j + i) = Suc (j + Suc i)" by auto
	      with Suc p show ?thesis by (simp only: su, blast)  
	    qed
	  qed
	  from this obtain l where contra: "p l \<in> set (concat ?cs) \<and> p (Suc l) \<in> set c" by auto
	  from this obtain d where d: "p l \<in> set d \<and> d \<in> set ?cs" by auto
	  from this obtain m where m: "m < length ?cs \<and> ?cs ! m = d" using set_conv_nth[where xs = ?cs] by force
	  with `?back (rc # cs)` d contra  have "\<not> g (p l, p (Suc l))"
            unfolding Pair by force
	  with Cons(2) show ?thesis by simp
	qed
      qed
    qed
  qed
qed

end

