(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Unraveling_Impl
imports
  Unraveling
  Conditional_Rewriting_Impl
begin

definition create_ctxts :: "('f, 'v) rules \<Rightarrow> (nat \<Rightarrow> ('f, 'v) ctxt) option"
where
  "create_ctxts R =
    (case R of
      [] \<Rightarrow> None
    | Cons l RR \<Rightarrow> do {
      cs \<leftarrow> Option_Monad.mapM (\<lambda>(l, r).
        (case l of
          Fun U (t # ts) \<Rightarrow> Some (More U [] \<box> ts)
        | _ \<Rightarrow> None)) RR;
      let n = length cs;
      Some (\<lambda>i. if i < length cs then cs ! i else \<box>) })"

definition
  create_U ::
    "(('f, 'v) crule \<times> ('f, 'v) rules) list \<Rightarrow> (('f, 'v) crule \<Rightarrow> nat \<Rightarrow> ('f, 'v) ctxt) option"
where
  "create_U c_rs = do {
    cr_ctxts \<leftarrow> Option_Monad.mapM (\<lambda> (cr, rs). do {
      guard (length rs = Suc (length (snd cr)));
      ctxt \<leftarrow> create_ctxts rs;
      Some (cr, ctxt)
    }) c_rs;
    let m = map_of cr_ctxts;
    Some (\<lambda> cr. case m cr of None \<Rightarrow> (\<lambda> _. \<box>) | Some ctxt \<Rightarrow> ctxt)
  }"

declare unraveling.lhs_n.simps[code] unraveling.rhs_n.simps[code]

definition
  rules_impl :: "(('f, 'v) crule \<Rightarrow> nat \<Rightarrow> ('f, 'v) ctxt) \<Rightarrow> ('f, 'v) crule \<Rightarrow> ('f, 'v) rules"
where
  "rules_impl U cr = map (\<lambda>i.
    (unraveling.lhs_n U cr i, unraveling.rhs_n U cr i)) [0 ..< Suc (length (snd cr))]"

lemma rules_impl[simp]: "set (rules_impl U cr) = unraveling.rules U cr"
proof -
  interpret unraveling "{}" U .
  obtain n where n: "Suc (length (snd cr)) = n" by auto
  have "{0..<Suc (length (snd cr))} = {i. i < Suc (length (snd cr))}" (is "?l = _")
    unfolding n by auto
  also have "... = {i. i \<le> length (snd cr)}" (is "_ = ?r") by auto
  finally have id: "?l = ?r" .
  show ?thesis unfolding rules_def rules_impl_def 
    by (simp del: upt_Suc add: id)
qed

definition check_unraveling :: "(('f :: show,'v :: show)crule \<times> ('f,'v)rules)list \<Rightarrow> ('f,'v)crules \<Rightarrow> ('f,'v)rules result"
  where "check_unraveling c_rs ctrs \<equiv> do {
      check_subseteq ctrs (map fst c_rs) <+? 
        (\<lambda> cr. shows ''did not find rule '' +@+ shows_crule cr +@+ shows_nl);
      U \<leftarrow> case create_U c_rs of None \<Rightarrow> error (shows ''unable to extract unraveling contexts'') | Some U \<Rightarrow> return U;
      check_allm (\<lambda> (c,rs). check (rules_impl U c = rs) (shows ''problem with rules of '' +@+ shows_crule c +@+ shows_nl)) c_rs;
      return (concat (map snd c_rs))
   } <+? (\<lambda> e. shows ''problem in unraveling '' +@+ shows_nl +@+ e)"

lemma check_unraveling:
  assumes ok: "check_unraveling c_rs ctrs = return r"
    and SN: "SN (rstep (set r))"
  shows "quasi_reductive (set ctrs)"
proof - 
  note ok = ok[unfolded check_unraveling_def, simplified]
  from ok obtain U where U: "create_U c_rs = Some U" by (cases "create_U c_rs", auto)
  note ok = ok[unfolded U, simplified]
  from ok have r: "r = concat (map snd c_rs)" by simp
  interpret unraveling "set ctrs" U .
  show ?thesis
  proof (rule SN_quasi_reductive, rule SN_subset[OF SN rstep_mono])
    show "UR \<subseteq> set r" 
    proof(rule subsetI)
      fix lr
      assume lr: "lr \<in> UR"
      from this[unfolded UR_def] obtain cr where cr: "cr \<in> set ctrs"
      and lr: "lr \<in> rules cr" by auto
      from ok cr obtain rs where mem: "(cr,rs) \<in> set c_rs" by auto
      from ok mem have "rules_impl U cr = rs" by auto
      from arg_cong[OF this, of set] have "rules cr = set rs" by simp
      with lr have lr: "lr \<in> set rs" by auto
      with mem show "lr \<in> set r" unfolding r by force
    qed
  qed
qed     
 

definition check_Z_vars :: "('f :: show,'v :: show) crules \<Rightarrow>  (('f, 'v) crule \<Rightarrow> nat \<Rightarrow> 'v list) \<Rightarrow> shows check"
where
  "check_Z_vars crs Z = do {
    check_allm 
     (\<lambda>cr .
      check_allm
       (\<lambda>i. do {
        check_subseteq (list_inter (X_impl cr i) (Y_impl cr i)) (Z cr i) 
         <+? (\<lambda>x. ''Variable '' +#+ shows x +@+ '' does not occur.'' +#+ shows_nl);
        check (distinct (Z cr i)) 
           ('' Z variables for rule are not distinct. '' +#+ shows_nl)
        } <+? (\<lambda>e. ''Conditions on Z variables  for '' +#+ shows_crule cr +@+ '' are violated.'' +#+ shows_nl +@+ e)
       ) [0..<length (snd cr)]
     ) crs 
  } <+? (\<lambda>e. ''The CTRS does not fulfill the condition on Z variables. '' +#+ shows_nl +@+ e)"

lemma check_Z_vars[simp]: 
 "isOK (check_Z_vars c_rs Z) = unraveling.Z_vars (set c_rs) Z"
proof-
 have x:"\<And>cr P. (\<forall>x\<in>set [0..<length (snd cr)]. P x) =  (\<forall>x <length (snd cr). P x) " by auto
 thus ?thesis
  unfolding check_Z_vars_def unraveling.Z_vars_def isOK_update_error isOK_forallM unfolding isOK_bind 
  isOK_update_error isOK_check_subseteq set_list_inter X_impl Y_impl isOK_check x by force
qed

definition create_zs :: "('f, 'v) rules \<Rightarrow>  (nat \<Rightarrow> 'v list) option"
where
  "create_zs R =
    (case R of
      [] \<Rightarrow> None
    | Cons l RR \<Rightarrow> do {
      cs \<leftarrow> Option_Monad.mapM (\<lambda>(l, r).
        (case l of
          Fun U (t # ts) \<Rightarrow> Some (map the_Var ts)
        | _ \<Rightarrow> None)) RR;
      Some (\<lambda>i. if i < length cs then (cs ! i) else [])})"

definition create_Z :: "(('f, 'v) crule \<times> ('f, 'v) rules) list \<Rightarrow> (('f, 'v) crule \<Rightarrow> nat \<Rightarrow> 'v list) option"
where "create_Z c_rs = do {
    cr_zs \<leftarrow> Option_Monad.mapM (\<lambda> (cr, rs). do {
      zs \<leftarrow> create_zs rs;
      Some (cr, zs)
    }) c_rs;
    let mc = map_of cr_zs;
    Some (\<lambda> cr. case mc cr of None \<Rightarrow> (\<lambda> _. []) | Some zs \<Rightarrow> zs)
  }"

definition create_Umap_cr :: "('f, 'v) crule \<Rightarrow> ('f, 'v) rules \<Rightarrow> (('f \<times> (('f, 'v) crule \<times> nat)) list) option"
where "create_Umap_cr cr R = (case R of 
       [] \<Rightarrow> None
     | Cons l RR \<Rightarrow> Option_Monad.mapM (\<lambda>((l, r), i).
        (case l of Fun U (t # ts) \<Rightarrow> Some (U,(cr,i)) | _ \<Rightarrow> None)) (zip RR [0..<length RR])) "

definition create_Umap :: "(('f, 'v) crule \<times> ('f, 'v) rules) list \<Rightarrow> ('f \<Rightarrow> (('f, 'v) crule \<times> nat) option)"
where "create_Umap c_rs = ( case Option_Monad.mapM (\<lambda> (cr, rs). create_Umap_cr cr rs) c_rs of 
 None \<Rightarrow> (\<lambda> _ .None) | Some u \<Rightarrow> map_of (concat u))"

abbreviation si :: "('f,'v) crule \<Rightarrow> nat =>  ('f,'v) term" where "si \<rho> i \<equiv> fst ((snd \<rho>)!i)"
abbreviation ti :: "('f,'v) crule \<Rightarrow> nat =>  ('f,'v) term" where "ti \<rho> i \<equiv> snd ((snd \<rho>)!i)"

definition check_prefix_equivalent :: "('f::show, 'v::show) crule \<Rightarrow> ('f, 'v) crule \<Rightarrow> nat \<Rightarrow> shows check"
 where "check_prefix_equivalent \<rho> \<rho>' n = do {
         check (n < length (snd \<rho>)) (''There are fewer than '' +#+ shows n +@+ '' conditions in '' +#+ shows_crule \<rho>);
         check (n < length (snd \<rho>')) (''There are fewer than '' +#+ shows n +@+ '' conditions in '' +#+ shows_crule \<rho>');
         check (fst (fst \<rho>) = fst (fst \<rho>')) (shows ''Left-hand sides are different.'');
         check_allm
          (\<lambda>i. check (ti \<rho> i = ti \<rho>' i) (''Rhs of conditions are different ''+#+ shows_nl)) 
          [0..<n];
         check_allm
          (\<lambda>i. check (si \<rho> i = si \<rho>' i) (''Lhs of conditions are different ''+#+ shows_nl)) 
          [0..<Suc n]
       }  <+? (\<lambda>e. ''Rules'' +#+ shows_crule \<rho> +@+ '' and '' +#+ shows_crule \<rho>' +@+ '' are not '' +#+ shows n +@+ '' equivalent.'' +#+ shows_nl +@+ e)"

lemma prefix_equivalent:
 "isOK(check_prefix_equivalent \<rho> \<rho>' n) = prefix_equivalent \<rho> \<rho>' n"
 unfolding prefix_equivalent_def check_prefix_equivalent_def  isOK_update_error
   isOK_bind isOK_check isOK_forallM set_upt by fastforce


definition check_f :: 
 "('f, 'v) crule \<Rightarrow> nat \<Rightarrow> 'f \<Rightarrow> ('f::show, 'v::show) crule list \<Rightarrow> (('f, 'v) crule \<Rightarrow> nat \<Rightarrow> ('f, 'v) ctxt)  \<Rightarrow> shows check"
 where "check_f cr' j f crs U =  
    check_allm 
     (\<lambda>cr.
      check_allm
       (\<lambda>i. 
        case U (cr:: ('f, 'v) crule) (i::nat) of More g b c aft \<Rightarrow> (
          if f = g then ( do{
           check (i = j) (''Same symbol occurs at different levels'' +#+ shows_nl);
           check_allm  (\<lambda>k. check (U cr k = U cr' k) (''Contexts are different ''+#+ shows_nl)) [0..< Suc j];
           check_prefix_equivalent cr cr' j}  <+? (\<lambda>e. ''Rules'' +#+ shows_crule cr +@+ '' and '' +#+ shows_crule cr' +@+ '' share a symbol.'' +#+ shows_nl +@+ e))
          else succeed)
         | _ \<Rightarrow> succeed
       ) [0..<length (snd cr)]
     ) crs"

lemma check_f: 
 "isOK(check_f \<rho> j f crs (U:: (('f, 'v) crule \<Rightarrow> nat \<Rightarrow> ('f::show, 'v::show) ctxt))) = 
   (\<forall> \<rho>' n' g b c a.  (\<rho>' \<in> (set crs) \<and> n' < length (snd \<rho>') \<and> U \<rho>' n' = More g b c a) \<longrightarrow> 
     f \<noteq> g \<or> (j = n' \<and> (\<forall>i \<le>j. U \<rho> i = U \<rho>' i) \<and> prefix_equivalent \<rho>' \<rho> j))"
proof-
   let ?check = "\<lambda> cr' j cr i . case U cr i of More g b c aft \<Rightarrow> (
          if f = g then ( do{
           check (i = j) (''Same symbol occurs at different levels'' +#+ shows_nl);
           check_allm  (\<lambda>k. check (U cr k = U cr' k) (''Contexts are different ''+#+ shows_nl)) [0..<Suc j];
           check_prefix_equivalent cr cr' j}  <+? (\<lambda>e. ''Rules'' +#+ shows_crule cr +@+ '' and '' +#+ shows_crule cr' +@+ '' share a symbol.'' +#+ shows_nl +@+ e))
          else succeed)
         | _ \<Rightarrow> succeed"
 { fix cr ::"('f,'v) crule" fix cr' i j
   let ?check = "?check cr' j cr i"
   have aux:"\<And>P j. (\<forall>x\<in>set [0..<Suc j]. P x) = (\<forall>x \<le> j. P x)" by auto
   have "isOK (?check) = (\<forall> g b c a. ( U cr i = More g b c a) \<longrightarrow> f \<noteq> g \<or> (j = i \<and> (\<forall>i \<le>j. U cr i = U cr' i) \<and> prefix_equivalent cr cr' j))"
   proof(cases "U cr i", simp)
    case (More g b c a) 
     thus ?thesis unfolding More ctxt.split isOK_error is_OK_if_return(2) 
      unfolding isOK_update_error isOK_bind isOK_check isOK_forallM prefix_equivalent aux by auto
   qed
 } note aux = this 
 have aux':"\<And>P j. (\<forall>x\<in>set [0..<j]. P x) = (\<forall>x < j. P x)" by auto
 {fix cr cr' j
  have "isOK(check_allm  (\<lambda>i. ?check cr' j cr i ) [0..<length (snd cr)]) =
  (\<forall> n' g b c a.  (n' < length (snd cr) \<and> U cr n' = More g b c a) \<longrightarrow> 
     f \<noteq> g \<or> (j = n' \<and> (\<forall>i \<le>j. U cr i = U cr' i) \<and> prefix_equivalent cr cr' j))" 
  unfolding isOK_update_error isOK_forallM aux aux' by blast
 } note * = this
 show ?thesis unfolding check_f_def isOK_update_error unfolding isOK_forallM * by metis
qed


definition check_U_cond :: 
 "(('f::show, 'v::show) crule \<Rightarrow> nat \<Rightarrow> ('f, 'v) ctxt) \<Rightarrow> ('f, 'v) crule list \<Rightarrow>  'f list \<Rightarrow> (('f, 'v) crule \<Rightarrow> nat \<Rightarrow> 'v list) \<Rightarrow> shows check"
 where "check_U_cond U crs F Z =  do {
    check_allm 
     (\<lambda>cr .
      check_allm
       (\<lambda>i. (case U cr i of 
             More f [] \<box> aft \<Rightarrow> do { 
             (*check (Umap f  = Some (cr,i)) (shows '' The internally generated  symbol map does not match. '');*)
             check_disjoint [f] F <+? (\<lambda>x. ''The function symbol '' +#+ show f +#+ '' is not fresh. '' +#+ shows_nl);
             check (aft = map Var (Z cr i)) (shows '' U does not map to Z vars '');
             check_f cr i f crs U
             }  <+? (\<lambda>e. ''Conditions for '' +#+ shows_crule cr +@+ '' at  '' +#+ shows i +@+ '' are violated.'' +#+ shows_nl +@+ e)
          | _ \<Rightarrow> error (shows '' Unexpected empty context.'')
)) [0..<length (snd cr)]
     ) crs 
  } <+? (\<lambda>e. ''The CTRS does not fulfill the condition on the U symbols.'' +#+ shows_nl +@+ e)"

lemma pe:"prefix_equivalent cr cr' n = prefix_equivalent cr' cr n"
 unfolding prefix_equivalent_def by fastforce

lemma check_U_cond[simp]:
 "isOK(check_U_cond U crs F Z) = U_cond U (set crs) (set F) Z" (is "?check = ?sem")
proof -
 let ?U = "\<lambda> \<rho> n. (\<exists> f. (U \<rho> n = (More f Nil Hole (map Var (Z \<rho> n)))  \<and> f \<notin> (set F) \<and>
    (\<forall> \<rho>' n' g b c a.  (\<rho>' \<in> (set crs) \<and> n' < length (snd \<rho>') \<and> U \<rho>' n' = More g b c a) \<longrightarrow> 
      f \<noteq> g \<or> (n = n' \<and> (\<forall>i \<le>n. U \<rho> i = U \<rho>' i) \<and> prefix_equivalent \<rho> \<rho>' n))))"
 let ?C = "\<lambda> cr i. case U cr i of 
             More f [] \<box> aft \<Rightarrow> do { 
             check_disjoint [f] F <+? (\<lambda>x. ''The function symbol '' +#+ show f +#+ '' is not fresh. '' +#+ shows_nl);
             check (aft = map Var (Z cr i)) (shows '' U does not map to Z vars '');
             check_f cr i f crs U
             }  <+?(\<lambda>e. ''Conditions for '' +#+ shows_crule cr +@+ '' at  '' +#+ shows i +@+ '' are violated.'' +#+ shows_nl +@+ e)
          | _ \<Rightarrow> error (shows '' Unexpected empty context.'')"
  show ?thesis
  proof
    assume ?check
    show ?sem
    proof (rule U_condI)
      fix cr i
      assume cr: "cr \<in> set crs" and i: "i < length (snd cr)"
      with `?check` have ok: "isOK(?C cr i)" unfolding check_U_cond_def by auto
      hence "\<exists>f b c a. U cr i = More f b c a" using isOK_error by (cases "U cr i", auto) 
      then obtain f b c a where 0:"U cr i = More f b c a" by blast
      from ok[unfolded 0] isOK_error have b:"b=[]" by (cases b, auto) 
      from ok[unfolded 0 b, simplified] isOK_error ctxt.exhaust have c:"c=\<box>" by (cases c, auto)
      note ok = ok[unfolded 0[unfolded b c], simplified]
      note 2 = conjunct1[OF conjunct2[OF ok]]
      from ok have a:"a = map Var (Z cr i)" by auto
      from conjunct2[OF conjunct2[OF ok], unfolded check_f] conjunct1[OF ok]
       0[unfolded a b c] 2 pe 
      show "?U cr i" by blast
    qed
  next
    assume ?sem
    show ?check unfolding check_U_cond_def isOK_forallM isOK_update_error set_upt
    proof (intro ballI)
      fix cr i
      assume cr: "cr \<in> set crs" and "i \<in> {0..<length (snd cr)}"
      hence i: "i < length (snd cr)" by auto
      let ?A = "\<lambda>f. U cr i = More f [] \<box> (map Var (Z cr i))"
      let ?C' = "\<lambda>f. f \<notin> set F"
      let ?D = "\<lambda>f. (\<forall> \<rho>' n' g b c a.  (\<rho>' \<in> (set crs) \<and> n' < length (snd \<rho>') \<and> U \<rho>' n' = More g b c a) \<longrightarrow> 
      f \<noteq> g \<or> (i = n' \<and> (\<forall>j \<le>i. U cr j = U \<rho>' j) \<and> prefix_equivalent cr \<rho>'  i))"
      from U_condD[OF `?sem` cr i] obtain f where "?A f \<and> ?C' f \<and> ?D f" by blast
      hence A:"?A f" and C:"?C' f" and D:"?D f" by (fast+)
      from D pe have 4:"isOK (check_f cr i f crs U)" unfolding check_f by blast
      from C 4 show "isOK (?C cr i)" unfolding A ctxt.case list.case by force
    qed      
  qed
qed

definition check_z_deterministic :: 
 "('f::show, 'v::show) crules \<Rightarrow> (('f, 'v) crule \<Rightarrow> nat \<Rightarrow> 'v :: show list)  \<Rightarrow> shows check"
 where "check_z_deterministic crs (Zv :: (('f, 'v) crule \<Rightarrow> nat \<Rightarrow> 'v list)) =  
    check_allm 
     (\<lambda>cr.
      check_allm
       (\<lambda>i. 
         check_subseteq (vars_term_list (fst (fst cr))) (Zv cr i)  <+? 
          (\<lambda> x. shows ''Some variable in lhs does not occur in Z_'' +@+ shows i +@+ shows ''. \<newline>'' ) 
         <+? (\<lambda>e. ''The unraveling is not Z-deterministic for rule '' +#+ shows_crule cr +@+ shows_nl +@+ e)
       ) [0..<length (snd cr)]
     ) crs"

lemma z_deterministic [simp]:
   "isOK (check_z_deterministic crs Z) = Z_deterministic (set crs) Z"
  unfolding Z_deterministic_def check_z_deterministic_def by force

definition check_z_det_unraveling :: "(('f :: show,'v :: show)crule \<times> ('f,'v)rules)list \<Rightarrow> ('f::show,'v::show)crules \<Rightarrow> ('f,'v)rules result"
where
  "check_z_det_unraveling c_rs ctrs = do {
    check_same_set ctrs (map fst c_rs) <+? (\<lambda> cr. shows ''did not find rule '' +@+ shows_crule cr +@+ shows_nl);
    U \<leftarrow> case create_U c_rs of None \<Rightarrow> error (shows ''unable to extract unraveling contexts'') | Some U \<Rightarrow> return U;
    Z \<leftarrow> case create_Z c_rs of None \<Rightarrow> error (shows ''unable to extract Z variables'') | Some Z \<Rightarrow> return Z;
    Umap \<leftarrow> return (create_Umap c_rs);
    check_U_cond U ctrs (funs_ctrs_list ctrs) Z;
    check_Z_vars ctrs Z;
    check_dctrs ctrs <+? (\<lambda> e. shows ''the CTRS is not a DCTRS '' +@+ shows_nl +@+ e);
    check_type3 ctrs <+? (\<lambda> e. shows ''the CTRS is not of type 3 '' +@+ shows_nl +@+ e);
    check_allm (\<lambda> (c,rs). check (rules_impl U c = rs) (shows ''problem with rules of '' +@+ shows_crule c +@+ shows_nl)) c_rs;
    check_left_linear_trs (concat (map snd c_rs)) <+? (\<lambda> e. shows ''the unraveled TRS is not left-linear '' +@+ shows_nl +@+ e);
    check_wf_ctrs ctrs <+? (\<lambda> e. shows ''the CTRS is not well-formed '' +@+ shows_nl +@+ e);
    check_z_deterministic ctrs Z <+? (\<lambda> e. shows ''unraveling is not Z-deterministic  '' +@+ shows_nl +@+ e);
    return (concat (map snd c_rs))
  } <+? (\<lambda>e. ''preconditions on the unraveling are not satisfied'' +#+ shows_nl +@+ e)"


lemma check_z_det_unraveling_CR:
  assumes ok: "check_z_det_unraveling c_rs ctrs = return  (ur :: ('f :: show, string) rules)"
    and inf: "infinite (UNIV :: 'f set)"
    and CR: "CR (rstep (set (ur :: ('f :: show, string) rules)))"
  shows "CR (cstep (set ctrs))"
proof -
 note ok = ok[unfolded check_z_det_unraveling_def, simplified]
 from ok obtain U where U: "create_U c_rs = Some U" by (cases "create_U c_rs", auto)
 from ok obtain Z where Z: "create_Z c_rs = Some Z" by (cases "create_Z c_rs", auto)
 note ok = ok[unfolded U Z, simplified]
 from ok have ur: "ur = concat (map snd c_rs)" by simp
 from ok z_deterministic have z:"Z_deterministic (set ctrs) Z" by auto
 from ok check_left_linear_trs have ll:"left_linear_trs (set ur)" by auto 
 from ok  have wf:"wf_ctrs (set ctrs)" by auto
 from ok have U_cond: "U_cond U (set ctrs) (funs_ctrs (set ctrs)) Z" by auto
 interpret unraveling "set ctrs" U .
 have ur1:"UR \<subseteq> set ur" 
 proof(rule subsetI)
      fix lr
      assume lr: "lr \<in> UR"
      from this[unfolded UR_def] obtain cr where cr: "cr \<in> set ctrs"
      and lr: "lr \<in> rules cr" by auto
      from ok cr obtain rs where mem: "(cr,rs) \<in> set c_rs" by auto
      from ok mem have "rules_impl U cr = rs" by auto
      from arg_cong[OF this, of set] have "rules cr = set rs" by simp
      with lr have lr: "lr \<in> set rs" by auto
      with mem show "lr \<in> set ur" unfolding ur by force
 qed
 have ur2:"UR \<supseteq> set ur" 
 proof(rule subsetI)
      fix lr
      assume lr: "lr \<in> set ur"
      from this[unfolded ur set_concat set_map] have "\<exists>(cr,rs) \<in> set c_rs. lr \<in> set rs" by fastforce
      then obtain cr rs where mem:"(cr,rs) \<in> set c_rs" "lr \<in> set rs" by blast
      with ok have a:"\<forall>x\<in>set c_rs. case x of (c, rs) \<Rightarrow> rules_impl U c = rs" by meson
      from imageI[OF mem(1), of fst] conjunct1[OF ok] have mem':"cr \<in> set ctrs" by simp
      from a[rule_format, OF mem(1), unfolded split] mem(2) have "lr \<in> rules cr" by force
      with mem' show "lr \<in> UR" unfolding UR_def by blast
 qed
 with ur1 have ur:"UR = set ur" by auto
 show ?thesis
 proof (rule Corollary_22)
   show "infinite (UNIV :: string set)" by (rule infinite_UNIV_listI)
   show "infinite (UNIV :: 'f set)" by (rule inf)
  qed (insert ur z ll CR wf ok, auto)
qed

hide_const (open) rules_impl create_U create_zs create_Z U_cond check_U_cond check_Z_vars 
 check_f si ti

end

